import pandas as pd
import os
from ..uniform_data_rule import ExternalServiceInterface
from ..uniform_data_rule import DATA_STORAGE as BASE_DATA_STORAGE

DATA_STORAGE = os.path.join(BASE_DATA_STORAGE, 'nfe')


class NfeSrvc(ExternalServiceInterface):

    @classmethod
    def get_from(cls, std, field, table=None, **options):
        file, options = cls._get_std_file(std)
        values = cls.read_table(file, field, **options)
        return {f'{std}': values}

    @classmethod
    def _get_std_file(cls, std):
        # TODO Improve organization
        return {
            'ncm': (
                'TABELA_NCM_2018-A.xlsx', dict(
                    skiprows=[0])),
            'unidade_comercial': (
                'Tabela_Unidades_de_Medida_Comercial_05072016.xls', dict(
                    skiprows=[0, 1]))
        }[std]

    @classmethod
    def read_table(cls, filename, field, table=None, sheet_num=0, header=[0], skiprows=None):
        """
         Returns a list with field from table
        """
        # TODO Other file extension
        # TODO More complicated types of tables
        df = pd.read_excel(os.path.join(DATA_STORAGE, filename), header=header, skiprows=skiprows)
        return df[field].values.tolist()



import unittest
from ..uniform_data_rule import MdmUniformType
import numpy as np


class TestValidate(unittest.TestCase):

    def test_validation_is_correct(self):
        rule = MdmUniformType(name='test', description='Test rule',
                       values={'custom': ['a', 'b', 'c']})
        data = np.array(['a', 'a', 'a', 'a', 'b', 'b', 'c'])
        success, logs = rule.validate(data)
        self.assertTrue(success)

        data = np.array(['a', 'a', 'a', 'a', 'b', 'b', 'd'])
        success, logs = rule.validate(data)
        self.assertFalse(success)
        self.assertEqual(logs['msg'], "Wrong data: ['d']")

        data = np.array(['a', 'a', 'a', 'a', 'b', 'b', 'd', 'e', 'f', 'g'])
        success, logs = rule.validate(data)
        self.assertFalse(success)
        self.assertEqual(logs['msg'], "Wrong data: ['d' 'e' 'f' 'g']")


""" Synchronize Data Model information on Carol

* Validate on code, what the code needs from Carol mappings

This module aims to maintain a synchronization with Carol of data model characteristics such as fields, types and
validation rules, as much as provide functionalities to organize data validation.

It works by getting from Carol all information from a data model and appending to it information from code. Some of the
validation rules, however, are not yet implemented on Carol. That's why we separate what should be updated on the MDM
and what should be only stored on code.


# TODO: Validation with external keys
# TODO: Study SQL Alchemy

"""
from ..data_models.data_model_build import DataModelBuild
import inspect

from .field import Field


class DataModel:
    _carol_name = None

    @classmethod
    def get_name(cls):
        """ Class name must be equal to field name

            E.g. Data Model with name 'dmname' should have a class named like:
                class DmName(DataModel):
                    ...
        """
        if cls._carol_name is None:
            return cls.__name__.lower()
        else:
            return cls._carol_name

    @classmethod
    def has_nested(cls):
        # TODO
        for field in cls.get_fields().values():
            if len(field.get_fields()) > 0:
                return True
        return False

    @classmethod
    def add_field(cls, field):
        """ Add Field to DataModel

        :param field: Field
        :return:
        """
        setattr(cls, field.__name__, field)

    @classmethod
    def from_carol(cls, carol):
        """ Get data from Carol to DataModel format

        :param carol_dm_dict: dictionary with this data model data from Carol

        :return: Data Model class with information from dm_dict
        """
        dm = DataModelBuild(carol).get_dm(cls.get_name())

        dm_name = dm['mdmName']

        # Create new DataModel object
        dm_class = type(dm_name.capitalize(), (DataModel,), {})

        # TODO: Get validations

        # TODO: Separate validations from fields

        # Transform all fields into classes, and all nested fields into subclasses (Dynamically)
        for field in dm['mdmFields']:
            validation = None
            f = Field.from_carol(field, validation)
            field_name = field['mdmName'].capitalize()
            setattr(dm_class, field_name, f)

        return dm_class

    @staticmethod
    def filter_metadata(dm_dict):
        """ Remove unused data from data model dict

        :param dm_dict: Dictionary with data model data from carol
        :return: Dictionary with filtered data model
        """

        data = {'mdmName': dm_dict['mdmDate'],
                'mdmFields': []}
        for field in dm_dict['mdmFields']:
            data['mdmFields'].update(Field.filter_metadata(field))
        return data

    @classmethod
    def _add_metadata(cls, carol, dm_dict, force_create=False, ignore_field_error=False):
        """ Add to class definition other information from those fields from Carol tenant

        :return:
        """
        # TODO

        fields_list = dm_dict.pop('mdmFields')
        data_model_name = dm_dict['mdmName']
        dm_carol = DataModelBuild(carol).get_dm(data_model_name)  # Get data from Carol
        dm_carol_fields = dm_carol.pop('mdmFields')

        # Merge
        dm_carol.update(dm_dict)
        dm_carol['mdmFields'] = []
        for field in fields_list:
            pos = cls._get_field_pos(field['mdmName'], dm_carol_fields)
            if pos == -1:
                if force_create:
                    new_field = Field.add_to_carol(carol, field)
                    # TODO Add to dm_carol_fields
                else:
                    if not ignore_field_error:
                        raise ValueError(f"Field {field['mdmName']} not found on Carol MDM")
            else:
                dm_carol_fields[pos].update(field)
                dm_carol['mdmFields'].append(dm_carol_fields[pos])

        return dm_carol

    @classmethod
    def get_validation_rules(cls):
        validations = {}
        for att_name in dir(cls):
            if att_name in ['validation_rule', 'uniform_rule', 'custom_validation']:
                validations.update({att_name: getattr(cls, att_name)})
        return validations

    @classmethod
    def get_fields(cls):
        """ Get Fields from this class

        :return: Dict of fields
        """
        fields = {}
        for att_name, att_val in cls.__dict__.items():
            if inspect.isclass(att_val):
                if issubclass(att_val, Field):
                    fields.update({att_name: att_val})
        return fields

    @classmethod
    def _get_field_pos(cls, field_name, fields_list):
        """ Get specific field's position from fields list

        :param field_name: Name of the desired field
        :param fields_list: Fields list
        :return: Field's position on list / -1 if not found
        """
        for pos in range(len(fields_list)):
            if field_name == fields_list[pos]['mdmName']:
                return pos
        return -1

    @classmethod
    def to_carol(cls, add_metadata=False, force_create=False):
        """ Set class to carol format

        :param add_metadata: Whether to include metadata from tenant or not
        :param force_create: If adds metadata, force field/data model/validation rule if not on Carol tenant

        :return: Dict with Data Model following Carol MDM format
        """
        # TODO
        carol_dm = {'mdmName': cls.get_name(),
                    'mdmFields': []}

        for att_name, att_val in cls.__dict__.items():
            if inspect.isclass(att_val):
                if issubclass(att_val, Field):
                    carol_dm['mdmFields'].append(att_val.to_carol(add_metadata=add_metadata, force_create=force_create))

        if add_metadata:
            return cls._add_metadata(carol_dm, force_create=force_create)

        return carol_dm

    @classmethod
    def create_validation_rules_from_fields(cls, carol=None, add_to_carol=False):
        """ Gather validation rules from all fields

        Since validation rules are exclusive to data models, we need a method to gather and create all to send to Carol

        :param carol: Carol object
        :param add_to_carol: Whether created rule should be added or not to carol
        :return: Success
        """
        rules = []
        dm_name = cls.get_name()
        for field in cls.get_fields().values():
            field_name = field.get_name()
            rules += [r.to_carol(dm_name, field_name, carol, add_to_carol) for r in field.get_nested_validation_rules()]

        return rules

    @classmethod
    def validate(cls, data, ignore_errors: bool=False):
        """ Validate if data follows Data Model definition

        :param data: Data Frame
        :return: Success, log
        """
        logs = {}
        success = True

        # TODO Validate data model format (fields)

        # Custom Validation
        if 'custom_validation' in cls.__dict__:
            # TODO Add msg as docstring
            custom_val_msg = 'Custom Validation - Check code.'
            custom_val_success = cls.custom_validation(data)
            success = success and custom_val_success
            logs.update({'custom_validation': {'success': custom_val_success,
                                               'msg': custom_val_msg}})

        # Run Fields validation
        logs['fields'] = {}
        for field_name, field in cls.get_fields().items():
            f_success, f_log = field.validate(data, ignore_errors=ignore_errors)
            success = success and f_success
            logs['fields'].update({field_name: f_log})

        return success, logs

    @classmethod
    def get_dtypes(cls):
        field_types = {}

        def _get_inner_fields(field, nested=None):
            field_name = field.get_name()
            if field.TYPE != Field.TYPE.NESTED:
                if nested is not None:
                    field_name = '_'.join(nested + [field_name])
                field_types.update({field_name: field.get_dtype()})
            else:
                if nested is None:
                    nested = [field_name]
                else:
                    nested.append(field_name)
                for f in field.get_fields().values():
                    _get_inner_fields(f, nested)

        for field in cls.get_fields().values():
            _get_inner_fields(field)

        return field_types

    @classmethod
    def update_version(cls, carol):
        #TODO
        pass


from ..luigi_extension import Task, WrapperTask
import luigi
from luigi import Parameter
import logging
import json
import os
import datetime
from ..apps import Apps
from ..carol import Carol
from ..staging import Staging

luigi.auto_namespace(scope=__name__)
logger = logging.getLogger(__name__)


class CarolAppConfig(luigi.Config):
    """ Task that contains configuration that might include parameters to load from Carol App

        1. Parameters are synchronized with Carol (back and forth)
        2. Documentation for app is generated from code and can be easily mantained/versioned.
        3. Single place to define all parameters / Less code on Run / facilitates logging/debugging/auditing

        app_carol: Dict with key:value parameters that came from Carol
        app_config: Dict with key:value parameters that are not from Carol
        app: Dict with key:value with all parameters

    """
    app_name = os.environ.get('CAROLAPPNAME')
    app_carol = None
    app = None

    @classmethod
    def initialization(cls):
        """ Define this function to do some logic on the parameters after they are extracted but before setting them
        on task. Parameters are saved on a variable named app.
        E.g.
            if cls.app['GROUP_ENTITY'] and cls.app['ENTITY'] != 'item':
                cls.app['ENTITY'] = 'short_mdmtaxid'
        """
        pass

    @classmethod
    def get_from_carol(cls, reset=False):
        """ Updates cls.app with values from Carol or Parameter object
        :return:
        """
        logger.debug('Getting Parameters data...')
        if cls.app is None:
            cls.app = {}

        for k, v in cls.get_params():
            if hasattr(v, 'carol') and v.carol:
                if cls.app_carol is None or reset:
                    login = Carol()
                    cls.app_carol = Apps(login)
                    cls.app_carol.get_settings(cls.app_name)
                    cls.app_carol = cls.app_carol.app_settings
                    for key, value in cls.app_carol.items():  # to avoid empty strings.
                        if value == '' or value is None:
                            cls.app_carol[key] = None
                        else:
                            try:
                                cls.app_carol[key] = json.loads(value)
                            except:
                                pass
                try:
                    if v.carol_name is not None:
                        logger.debug(f'{v.carol_name}: {cls.app_carol[v.carol_name]}')
                        carol_val = cls.app_carol[v.carol_name]
                    else:
                        carol_val = cls.app_carol[k]
                    # Convert Parameter type
                    if carol_val is not None:
                        if isinstance(v, luigi.DateParameter):
                            carol_val = datetime.datetime.strptime(carol_val, '%Y-%m-%d')  # TEMP - Carol should have a Date Parameter type
                    v._default = carol_val
                except KeyError as e:
                    logger.warning(f"Could not set up variable from Carol. Key = {str(e)}")
            cls.app.update({k: v})

        cls.initialization()
        return cls

    def to_carol(self):
        # TODO: Updates carol app with class parameters
        pass


class inherits_carol(object):
    """ Add CarolConfig parameters to a Task
    """

    def __init__(self, params_task):
        self.params_task = params_task

    def __call__(self, task_that_inherits):
        self.params_task.get_from_carol()
        for name, val in self.params_task.app.items():
                setattr(task_that_inherits, name, val)
        return task_that_inherits

# Create Luigi mappings


class StagingIngestion(Task):
    """ Task to execute Staging ingestion
    """

    tenant = Parameter()
    connector_name = Parameter()
    staging_name = Parameter()
    return_dask_graph = Parameter(significant=False, default=False)
    cols = Parameter(default=[])

    def easy_run(self, inputs):
        login = Carol()
        stag = Staging(login)
        logger.debug(f'Executing parquet query for staging {self.staging_name}')
        return stag.fetch_parquet(staging_name=self.staging_name, connector_name=self.connector_name,
                                  backend='pandas', return_dask_graph=self.return_dask_graph,
                                  columns=list(self.cols), merge_records=True)


class DataModelValidation(Task):
    """ Task to execute Data Model validation

    """
    tenant = Parameter(default='Unspecified')
    ignore_errors = Parameter(default=True)

    def easy_run(self, inputs):
        log = {'domain': self.tenant,
               'datamodel': self.dm.get_name()}

        success, log_dm = self.dm.validate(inputs[0], ignore_errors=self.ignore_errors)

        if not self.ignore_errors and not success:
            raise ValueError(log)
        log['log'] = log_dm
        return log



from .utils.singleton import KeySingleton
from .apps import Apps


class Settings(metaclass=KeySingleton):
    def __init__(self, carol):
        self.carol = carol
        self.app = Apps(carol).get_settings(app_name=self.carol.app_name)

    def get(self, key):
        return self.app.appSettings[key]

    def get_full(self, key):
        return self.app.fullSettings[key]

    def all(self):
        return self.app.appSettings

    def all_full(self):
        return self.app.fullSettings




import unittest
from ..data_model import DataModel, Field


class DataModelMock(DataModel):
    class Nestedone(Field):
        TYPE = Field.TYPE.NESTED

        class Nestedonedeep(Field):
            TYPE = Field.TYPE.LONG


class TestLuigiTasks(unittest.TestCase):
    def test_validation_tasks_are_created(self):
        pass



from string import Formatter


class Carolina:
    token = None

    def __init__(self, carol):
        self.carol = carol
        self.legacy_mode = carol.legacy_mode
        self.legacy_bucket = carol.legacy_bucket
        self.client = None
        self.engine = None
        self.token = None
        self.cds_app_storage_path = None
        self.cds_golden_path = None
        self.cds_staging_path = None
        self.cds_staging_master_path = None
        self.cds_staging_rejected_path = None


    def _legacy_mode(self):
        response = self.carol.call_api('v1/carolina/carolina/token', params={'carolAppName': self.carol.app_name})

        token = {}
        token['tenant_name'] = self.carol.tenant['mdmName']
        token['engine'] = "AWS-S3"
        token['cdsAppStoragePath'] = {
            "bucket": self.legacy_bucket,
            "path": f"storage/{self.carol.tenant['mdmId']}/{self.carol.app_name}/files"
        }
        token['cdsGoldenPath'] = {
            "bucket": self.legacy_bucket,
            "path": f"carol_export/{self.carol.tenant['mdmId']}/{{dm_name}}/golden"
        }
        token['cdsStagingPath'] = {
            "bucket": self.legacy_bucket,
            "path": f"carol_export/{self.carol.tenant['mdmId']}/{{connector_id}}_{{staging_type}}/staging"
        }
        token['cdsStagingMasterPath'] = {
            "bucket": self.legacy_bucket,
            "path": f"carol_export/{self.carol.tenant['mdmId']}/{{connector_id}}_{{staging_type}}/master_staging"
        }
        token['cdsStagingRejectedPath'] = {
            "bucket": self.legacy_bucket,
            "path": f"carol_export/{self.carol.tenant['mdmId']}/{{connector_id}}_{{staging_type}}/rejected_staging"
        }

        token['aiAccessKeyId'] = response['aiAccessKeyId']
        token['aiSecretKey'] = response['aiSecretKey']
        token['aiAccessToken'] = response['aiAccessToken']
        token['aiTokenExpirationDate'] = response['aiTokenExpirationDate']


        return token

    def init_if_needed(self):
        if self.client:
            return

        if Carolina.token is None:
            if self.legacy_mode:
                token = self._legacy_mode()
            else:
                token = self.carol.call_api('v1/storage/storage/token', params={'carolAppName': self.carol.app_name})
                token['tenant_name'] = self.carol.tenant['mdmName']
            Carolina.token = token
        elif Carolina.token.get('tenant_name', '') != self.carol.tenant['mdmName']:
            if self.legacy_mode:
                token = self._legacy_mode()
            else:
                token = self.carol.call_api('v1/storage/storage/token', params={'carolAppName': self.carol.app_name})
                token['tenant_name'] = self.carol.tenant['mdmName']
            Carolina.token = token

        token = Carolina.token
        self.engine = token['engine']

        self.cds_app_storage_path = token['cdsAppStoragePath']
        self.cds_golden_path = token['cdsGoldenPath']
        self.cds_staging_path = token['cdsStagingPath']
        self.cds_staging_master_path = token['cdsStagingMasterPath']
        self.cds_staging_rejected_path = token['cdsStagingRejectedPath']

        if self.engine == 'GCP-CS':
            self._init_gcp(token)
        elif self.engine == 'AWS-S3':
            self._init_aws(token)

    def _init_gcp(self, token):
        from google.oauth2 import service_account
        from google.cloud import storage

        self.token = token['token']
        gcp_credentials = service_account.Credentials.from_service_account_info(token['token'])
        self.client = storage.Client(credentials=gcp_credentials, project=token['token']['project_id'])

    def _init_aws(self, token):
        import boto3
        self.token = token
        ai_access_key_id = token['aiAccessKeyId']
        ai_secret_key = token['aiSecretKey']
        ai_access_token = token['aiAccessToken']
        ai_token_expiration_date = token['aiTokenExpirationDate']

        boto3.setup_default_session(aws_access_key_id=ai_access_key_id, aws_secret_access_key=ai_secret_key,
                                    aws_session_token=ai_access_token)
        self.client = boto3.resource('s3')

    def get_client(self):
        self.init_if_needed()
        return self.client

    def get_bucket_name(self, space):
        if space == 'golden':
            template = self.cds_golden_path['bucket']
        elif space == 'staging':
            template = self.cds_staging_path['bucket']
        elif space == 'staging_master':
            template = self.cds_staging_master_path['bucket']
        elif space == 'staging_rejected':
            template = self.cds_staging_rejected_path['bucket']

        name = Formatter().vformat(template, None, {'tenant_id': self.carol.tenant['mdmId']})
        return name

    def get_path(self, space, vars):
        vars['tenant_id'] = self.carol.tenant['mdmId']

        if space == 'golden':
            template = self.cds_golden_path['path']
        elif space == 'staging':
            template = self.cds_staging_path['path']
        elif space == 'staging_master':
            template = self.cds_staging_master_path['path']
        elif space == 'staging_rejected':
            template = self.cds_staging_rejected_path['path']

        name = Formatter().vformat(template, None, vars)
        return name



import os
import tempfile

__version__ = '2.13.6'


__BUCKET_NAME__= 'carol-internal'
__TEMP_STORAGE__ = os.path.join(tempfile.gettempdir(),'carolina/cache')

__CONNECTOR_PYCAROL__ = 'f9953f6645f449baaccd16ab462f9b64'




import os
import pickle
import gzip
import pandas as pd
from .. import __TEMP_STORAGE__


class StorageGCPCS:
    def __init__(self, carol, carolina):
        self.carol = carol
        self.carolina = carolina
        self.client = None
        self.bucket_app_storage = None

    def _init_if_needed(self):
        if self.client is not None:
            return

        self.client = self.carolina.get_client()
        self.bucket_app_storage = self.client.bucket(self.carolina.cds_app_storage_path['bucket'])
        if not os.path.exists(__TEMP_STORAGE__):
            os.makedirs(__TEMP_STORAGE__)

    def save(self, name, obj, format='pickle', parquet=False, cache=True):
        self._init_if_needed()
        remote_file_name = f"{self.carolina.cds_app_storage_path['path']}/{name}"
        local_file_name = os.path.join(__TEMP_STORAGE__, remote_file_name.replace("/", "-"))

        bucket = self.bucket_app_storage

        if parquet:
            if not isinstance(obj, pd.DataFrame):
                raise ValueError("Object to be saved as parquet must be a DataFrame")
            obj.to_parquet(local_file_name)
        elif format == 'joblib':
            import joblib

            if not cache:
                from io import BytesIO
                with BytesIO() as buffer:
                    joblib.dump(obj, buffer)
                    buffer.seek(0)
                    bucket.upload_fileobj(buffer, remote_file_name)
                return
            else:
                joblib.dump(obj, local_file_name)
        elif format == 'pickle':
            with gzip.open(local_file_name, 'wb') as f:
                pickle.dump(obj, f, pickle.HIGHEST_PROTOCOL)
        elif format == 'file':
            local_file_name = obj
        else:
            raise ValueError("Supported formats are pickle, joblib or file")

        blob = bucket.blob(remote_file_name)

        blob.upload_from_filename(filename=local_file_name)
        os.utime(local_file_name, None)

    def load(self, name, format='pickle', parquet=False, cache=True, storage_space='app_storage', columns=None):
        self._init_if_needed()
        if storage_space == 'app_storage':
            remote_file_name = f"{self.carolina.cds_app_storage_path['path']}/{name}"
            bucket = self.bucket_app_storage
        else:
            remote_file_name = name
            bucket = self.client.bucket(self.carolina.get_bucket_name(storage_space))

        local_file_name = os.path.join(__TEMP_STORAGE__, remote_file_name.replace("/", "-"))

        has_cache = cache and os.path.isfile(local_file_name)

        if has_cache:
            localts = os.stat(local_file_name).st_mtime
        else:
            localts = 0

        blob = bucket.blob(remote_file_name)
        if not blob.exists():
            return None

        blob.reload()
        remote_ts = blob.updated.timestamp()

        if not cache and format == 'raw':
            import joblib
            from io import BytesIO
            buffer = BytesIO()
            blob.download_to_file(buffer)
            return buffer

        elif not cache and format == 'joblib':
            import joblib
            from io import BytesIO
            with BytesIO() as buffer:
                blob.download_to_file(buffer)
                return joblib.load(buffer)

        # Local cache is outdated
        if localts < remote_ts:
            blob.download_to_filename(local_file_name)

        if os.path.isfile(local_file_name):
            if parquet:
                return pd.read_parquet(local_file_name, columns=columns)
            elif format == 'joblib':
                import joblib
                return joblib.load(local_file_name)
            elif format == 'pickle':
                with gzip.open(local_file_name, 'rb') as f:
                    return pickle.load(f)
            elif format == 'file':
                return local_file_name
            else:
                raise ValueError("Supported formats are pickle, joblib or file")
        else:
            return None

    def exists(self, name):
        self._init_if_needed()
        remote_file_name = f"{self.carolina.cds_app_storage_path['path']}/{name}"
        print(remote_file_name)

        blob = self.bucket_app_storage.blob(remote_file_name)
        return blob.exists()

    def delete(self, name):
        self._init_if_needed()
        remote_file_name = f"{self.carolina.cds_app_storage_path['path']}/{name}"

        blob = self.bucket_app_storage.blob(remote_file_name)
        blob.delete()

        local_file_name = os.path.join(__TEMP_STORAGE__, remote_file_name.replace("/", "-"))
        if os.path.isfile(local_file_name):
            os.remove(local_file_name)

    def build_url_parquet_golden(self, dm_name):
        path = self.carolina.get_path("golden", {'dm_name': dm_name})
        return f'gcs://{self.carolina.get_bucket_name("golden")}/{path}'

    def build_url_parquet_staging(self, staging_name, connector_id):
        self._init_if_needed()
        path = self.carolina.get_path("staging", {'connector_id': connector_id, 'staging_type': staging_name})
        bucket_name = self.carolina.get_bucket_name("staging")
        bucket = self.client.bucket(bucket_name)
        blobs = list(bucket.list_blobs(prefix=path))
        if len(blobs) == 0:
            return None
        return f'gcs://{bucket_name}/{path}'

    def build_url_parquet_staging_master(self, staging_name, connector_id):
        self._init_if_needed()
        path = self.carolina.get_path("staging_master", {'connector_id': connector_id, 'staging_type': staging_name})
        bucket_name = self.carolina.get_bucket_name("staging_master")
        bucket = self.client.bucket(bucket_name)
        blobs = list(bucket.list_blobs(prefix=path))
        if len(blobs) == 0:
            return None
        return f'gcs://{bucket_name}/{path}'

    def build_url_parquet_staging_rejected(self, staging_name, connector_id):
        self._init_if_needed()
        path = self.carolina.get_path("staging_rejected", {'connector_id': connector_id, 'staging_type': staging_name})
        bucket_name = self.carolina.get_bucket_name("staging_rejected")
        bucket = self.client.bucket(bucket_name)
        blobs = list(bucket.list_blobs(prefix=path))
        if len(blobs) == 0:
            return None
        return f'gcs://{bucket_name}/{path}'

    def get_dask_options(self):
        return {'token': self.carolina.token}

    def get_golden_file_paths(self, dm_name):
        self._init_if_needed()
        bucket = self.client.bucket(self.carolina.get_bucket_name('golden'))
        path = self.carolina.get_path('golden', {'dm_name': dm_name})
        blobs = bucket.list_blobs(prefix=path, delimiter=None)
        return [{'storage_space': 'golden', 'name': i.name} for i in blobs if i.name.endswith('.parquet')]

    def get_staging_file_paths(self, staging_name, connector_id):
        self._init_if_needed()

        bucket_staging = self.client.bucket(self.carolina.get_bucket_name('staging'))
        path_staging = self.carolina.get_path("staging", {'connector_id': connector_id, 'staging_type': staging_name})
        blobs_staging = list(bucket_staging.list_blobs(prefix=path_staging))

        bucket_master = self.client.bucket(self.carolina.get_bucket_name('staging_master'))
        path_master = self.carolina.get_path("staging_master", {'connector_id': connector_id, 'staging_type': staging_name})
        blobs_master = list(bucket_master.list_blobs(prefix=path_master))

        bucket_rejected = self.client.bucket(self.carolina.get_bucket_name('staging_rejected'))
        path_rejected = self.carolina.get_path("staging_rejected", {'connector_id': connector_id, 'staging_type': staging_name})
        blobs_rejected = list(bucket_rejected.list_blobs(prefix=path_rejected))

        bs = [{'storage_space': 'staging', 'name': i.name} for i in blobs_staging if i.name.endswith('.parquet')]
        bm = [{'storage_space': 'staging_master', 'name': i.name} for i in blobs_master if i.name.endswith('.parquet')]
        br = [{'storage_space': 'staging_rejected', 'name': i.name} for i in blobs_rejected if i.name.endswith('.parquet')]

        ball = bs + bm + br
        return ball



class Singleton(type):
    _instance = None

    def __call__(cls, *args, **kwargs):
        if cls._instance is None:
            cls._instance = super(Singleton, cls).__call__(*args, **kwargs)
        return cls._instance


class KeySingleton(type):
    _instances = {}

    def __call__(cls, *args, **kwargs):
        if len(args) == 0:
            key = ''
        else:
            key = args[0]
        if key not in cls._instances:
            cls._instances[key] = super(KeySingleton, cls).__call__(*args, **kwargs)
        return cls._instances[key]




import unittest
import pandas as pd
from ..field import Field
from ..validation_rule.validation_rule import ValidationRule
from ..validation_rule.uniform_data_rule import MdmUniformType
from unittest.mock import patch


class Fieldmock(Field):
    TYPE = Field.TYPE.NESTED

    class Nestedone(Field):
        TYPE = Field.TYPE.NESTED

        class Nestedonedeep(Field):
            TYPE = Field.TYPE.LONG

            @classmethod
            def custom_validation(cls, data):
                # Data must be greater than 5
                return not any(data > 5)

    class Nestedtwo(Field):
        TYPE = Field.TYPE.DOUBLE

        @classmethod
        def uniform_rule(cls):
            return MdmUniformType.get_rule('test')

    class Nestedthree(Field):
        TYPE = Field.TYPE.DOUBLE

        @classmethod
        def validation_rule(cls):
            return ValidationRule(ValidationRule.OPERATION.EQUALS, [1])


fieldmock_carol_code = {
            'mdmName':'fieldmock',
            'mdmMappingDataType':Field.TYPE.NESTED,
            'mdmFields': [
                 {
                     'mdmName': 'nestedone',
                     'mdmMappingDataType': Field.TYPE.NESTED,
                     'mdmFields':[
                         {
                             'mdmName':'nestedonedeep',
                             'mdmMappingDataType':Field.TYPE.LONG,
                             'mdmFields': []
                         }
                     ]
                 },
                 {
                     'mdmName': 'nestedtwo',
                     'mdmMappingDataType': Field.TYPE.DOUBLE,
                     'mdmFields': []
                 },
                {
                    'mdmName': 'nestedthree',
                    'mdmMappingDataType': Field.TYPE.DOUBLE,
                    'mdmFields': []
                }
             ]
            }


class TestField(unittest.TestCase):

    def test_interpret_fields_defined_on_code(self):
        self.assertDictEqual(Fieldmock.to_carol(), fieldmock_carol_code)

    def test_interpret_fields_defined_at_carol(self):
        FromCarol = Field.from_carol(fieldmock_carol_code)
        self.assertEqual(set(FromCarol.__dict__.keys()), set(Fieldmock.__dict__.keys()))
        self.assertEqual(set(FromCarol.Nestedone.__dict__.keys()), set(Fieldmock.Nestedone.__dict__.keys()))


    def test_get_fields(self):
        self.assertEqual(set(Fieldmock.get_fields().values()), {Fieldmock.Nestedone, Fieldmock.Nestedtwo,
                                                                 Fieldmock.Nestedthree})

    def test_get_field_parent(self):
        parent = Fieldmock.Nestedone.Nestedonedeep.get_parent()
        self.assertEquals(parent, Fieldmock.Nestedone)


class TestFieldValidation(unittest.TestCase):

    def test_type_validation(self):
        data = pd.DataFrame({'nestedonedeep': ['a', 'b', 'c', 'd', 'e', 'f']})
        success, logs = Fieldmock.Nestedone.Nestedonedeep.validate(data, ignore_errors=True)
        self.assertFalse(success)
        data = pd.DataFrame({'nestedonedeep': [-1, 0, 1, 2, 3, 4]})
        success, logs = Fieldmock.Nestedone.Nestedonedeep.validate(data, ignore_errors=False)
        self.assertTrue(success)

    def test_custom_validation(self):
        data = pd.DataFrame({'nestedonedeep': [1, 2, 3, 4, 5, 6]})
        success, logs = Fieldmock.Nestedone.Nestedonedeep.validate(data, ignore_errors=True)
        self.assertFalse(success)
        data = pd.DataFrame({'nestedonedeep': [-1, 0, 1, 2, 3, 4]})
        success, logs = Fieldmock.Nestedone.Nestedonedeep.validate(data, ignore_errors=False)
        self.assertTrue(success)

    def test_validation_rule_validation(self):
        data = pd.DataFrame({'nestedthree': [1., 2., 3., 4., 5., 6.]})
        success, logs = Fieldmock.Nestedthree.validate(data, ignore_errors=True)
        self.assertFalse(success)
        data = pd.DataFrame({'nestedthree': [1., 1., 1., 1., 1., 1.]})
        success, logs = Fieldmock.Nestedthree.validate(data, ignore_errors=False)
        self.assertTrue(success)

    @patch('pycarol.pipeline.validation_rule.uniform_data_rule.MdmUniformType.get_rule',
           return_value=MdmUniformType('test', 'Test description', version='1.0.0', values={'a': [1., 2., 3.]}))
    def test_uniform_rule_validation(self, mocked):
        data = pd.DataFrame({'nestedtwo': [1., 2., 3., 4., 5., 6.]})
        success, logs = Fieldmock.Nestedtwo.validate(data, ignore_errors=True)
        self.assertFalse(success)
        data = pd.DataFrame({'nestedtwo': [1., 1., 2., 2., 3., 3.]})
        success, logs = Fieldmock.Nestedtwo.validate(data, ignore_errors=False)
        self.assertTrue(success)

    @patch('pycarol.pipeline.validation_rule.uniform_data_rule.MdmUniformType.get_rule',
           return_value=MdmUniformType('test', 'Test description', version='1.0.0', values={'a': [1., 2., 3.]}))
    def test_all_validation(self, mocked):
        data = pd.DataFrame({
            'fieldmock_nestedone_nestedonedeep':['a', 'b', 'c', 'd', 'e', 'f'],
            'fieldmock_nestedtwo': [1, 2, 3, 4, 5, 6],
            'fieldmock_nestedthree': [1., 2., 3., 4., 5., 6.]
        })
        success, logs = Fieldmock.validate(data, ignore_errors=True)
        self.assertFalse(success)

        data = pd.DataFrame({
            'nestedone_nestedonedeep': [1, 2, 3, 4, 5, 6],
            'nestedtwo': [1., 1., 2., 2., 3., 3.],
            'nestedthree': [1., 1., 1., 1., 1., 1.]
        })
        success, logs = Fieldmock.Nestedtwo.validate(data, ignore_errors=False)
        self.assertTrue(success)

    # TODO: Test logs



# TEMP: This task should be placed somewhere else
from ..uniform_data_rule import MdmUniformType
from ..services.ISO import IsoSrvc
from ..services.NFe import NfeSrvc


def generate_rules():
    """ Create new uniform rules that are used in the Supply Chain Data models
    OBS! *************
    This method is intended to be temporary, since data model uniform rules should be created on the Carol platform
    Check MDMUniformType for more details.
    """
    # Currency: Generate from ISO 4217 and table Cryptocurrencies
    currency_rule = MdmUniformType(name='currency', description='Default Format for Currencies')
    currency_rule.get_values_from('4217', 'Ccy', table='CcyNtry', service=IsoSrvc).post()

    # Commercial Unity

    # NCM
    ncm_rule = MdmUniformType(name='ncm', description='NCM Identifiers')
    ncm_rule.get_values_from('ncm', 'NCM', service=NfeSrvc).post()

    # Purchase Status
    MdmUniformType(name='purchase_status', description='Status of a purchase',
                   values={'custom':['printed', 'open', 'cancelled', 'delivered']}).post()

    # Purchase Type
    MdmUniformType(name='purchase_type', description='Type of purchase',
                   values={'custom':['invoice', 'order', 'other']})

    # Government Id Type
    MdmUniformType(name='government_id', description='Government ID',
                   values=['invoice', 'order', 'other'])




from urllib3.util.retry import Retry
import requests
from requests.adapters import HTTPAdapter
import json
import os
import os.path
from .auth.ApiKeyAuth import ApiKeyAuth
from .auth.PwdAuth import PwdAuth
from .tenant import Tenant
from . import __CONNECTOR_PYCAROL__


class Carol:
    def __init__(self, domain=None, app_name=None, auth=None, connector_id=None, port=443, verbose=False):
        """
        This class handle all Carol`s API calls
        It will handle all API calls, for a given authentication method.
        :param domain: `str`
            Teanant name
        :param app_name: `str`
            Carol app name.
        :param auth: `PwdAuth` or `ApiKeyAuth` object
            Auth Carol object to handle authentication
        :param connector_id: `str`, default `__CONNECTOR_PYCAROL__`
            Connector Id
        :param port: `int`, default 443
            Port to be used (when running locally)
        :param verbose: `bool`, default `False
            If True will print the header, method and URL of each API call.
        """
        self.legacy_mode = False
        self.legacy_bucket = 'carol-internal'

        settings = dict()
        if auth is None and domain is None:

            domain = os.getenv('CAROLTENANT')
            app_name = os.getenv('CAROLAPPNAME')
            auth_token = os.getenv('CAROLAPPOAUTH')
            connector_id = os.getenv('CAROLCONNECTORID')
            assert (domain is not None) and (app_name is not None) and (auth_token is not None) and (connector_id is not None),\
                    "One of the following env variables are missing:\n " \
                    f"CAROLTENANT: {domain}\nCAROLAPPNAME: {app_name}\nCAROLAPPOAUTH: {auth}\nCAROLCONNECTORID: {connector_id}\n"
            auth = ApiKeyAuth(auth_token)


        if connector_id is None:
            connector_id =  __CONNECTOR_PYCAROL__

        if domain is None or app_name is None or auth is None:
            raise ValueError("domain, app_name and auth must be specified as parameters, in the app_config.json file " +
                             "or in the environment variables CAROLTENANT, CAROLAPPNAME, CAROLAPPOAUTH OR CAROLUSER+CAROLPWD and " +
                             "CAROLCONNECTORID")

        self.domain = domain
        self.app_name = app_name
        self.port = port
        self.verbose = verbose
        self.tenant = Tenant(self).get_tenant_by_domain(domain)
        self.connector_id = connector_id
        self.auth = auth
        self.auth.set_connector_id(self.connector_id)
        self.auth.login(self)
        self.response = None


    def build_ws_url(self, path):
        """
        Return the URS to be used with websocket and queries.
        :param path: `str`
            API and point.
        :return: str
        """
        return 'wss://{}.carol.ai:{}/websocket/{}'.format(self.domain, self.port, path)

    @staticmethod
    def _retry_session(retries=5, session=None, backoff_factor=0.5, status_forcelist=(500, 502, 503, 504, 524),
                       method_whitelist=frozenset(['HEAD', 'TRACE', 'GET', 'PUT', 'OPTIONS', 'DELETE'])):
        """
        :param retries: `int`, default `5`
            Number of retries for the API all
        :param session: Session object defaut `None`
            It allows you to persist certain parameters across requests.
        :param backoff_factor: `float`, default `0.5`
            Backoff factor to apply between attempts. It will
            sleep for: {backoff factor} * (2 ^ ({retries} - 1)) seconds
        :param status_forcelist: `iterable`, default (500, 502, 503, 504, 524)
            A set of integer HTTP status codes that we should force a retry on.
             A retry is initiated if the request method is in method_whitelist and the response status code is in status_forcelist.
        :param method_whitelist: `iterable`, default frozenset(['HEAD', 'TRACE', 'GET', 'PUT', 'OPTIONS', 'DELETE']))
            Set of uppercased HTTP method verbs that we should retry on.
        :return: session object
        """
        session = session or requests.Session()
        retry = Retry(
            total=retries,
            read=retries,
            connect=retries,
            backoff_factor=backoff_factor,
            status_forcelist=status_forcelist,
            method_whitelist=method_whitelist,
        )
        adapter = HTTPAdapter(max_retries=retry)
        session.mount('http://', adapter)
        session.mount('https://', adapter)
        return session

    def call_api(self, path, method=None, data=None, auth=True, params=None, content_type='application/json',retries=5,
                 session=None, backoff_factor=0.5, status_forcelist=(500, 502, 503, 504, 524), downloadable=False,
                 method_whitelist=frozenset(['HEAD', 'TRACE', 'GET', 'PUT', 'OPTIONS', 'DELETE']), errors='raise',
                 extra_headers=None,
                 **kwds):
        """
        :param path:
        :param method:
        :param data:
        :param auth:
        :param params:
        :param content_type:
        :param retries:
        :param session:
        :param backoff_factor:
        :param status_forcelist:
        :param downloadable:
        :param method_whitelist:
        :param extra_headers:
        :param errors : {‘ignore’, ‘raise’}, default ‘raise’
                If ‘raise’, then invalid request will raise an exception
                If ‘ignore’, then invalid request will return the request response
        :param kwds:
        :return:
        """

        extra_headers = extra_headers or {}
        url = 'https://{}.carol.ai:{}/api/{}'.format(self.domain, self.port, path)

        if method is None:
            if data is None:
                method = 'GET'
            else:
                method = 'POST'

        headers = {'accept': 'application/json'}
        if auth:
            self.auth.authenticate_request(headers)

        data_json = None
        if method == 'GET':
            pass

        elif (method == 'POST') or (method == 'DELETE') or (method == 'PUT'):
            headers['content-type'] = content_type

            if content_type == 'application/json':
                data_json = data
                data = None

        headers.update(extra_headers)
        __count = 0
        while True:
            if session is None:
                session = self._retry_session(retries=retries, session=session, backoff_factor=backoff_factor,
                                              status_forcelist=status_forcelist, method_whitelist=method_whitelist)

            response = session.request(method=method, url=url, data=data, json=data_json,
                                       headers=headers, params=params, **kwds)

            if self.verbose:
                if data_json is not None:
                    print("Calling {} {}. Payload: {}. Params: {}".format(method, url, data_json, params))
                else:
                    print("Calling {} {}. Payload: {}. Params: {}".format(method, url, data, params))
                print("        Headers: {}".format(headers))

            if response.ok or errors == 'ignore':
                if downloadable:  #Used when downloading carol app file.
                    return response

                response.encoding = 'utf-8'
                self.response = response
                if response.text == '':
                    return {}
                return json.loads(response.text)

            elif (response.reason == 'Unauthorized') and  isinstance(self.auth,PwdAuth):
                self.auth.get_access_token()  #It will refresh token if Unauthorized
                __count+=1
                if __count<5: #To avoid infinity loops
                    continue
                else:
                    raise Exception('Too many retries to refresh token.\n',response.text)
            raise Exception(response.text)

    def issue_api_key(self):
        resp = self.call_api('v2/apiKey/issue', data={
            'connectorId': self.connector_id
        }, content_type='application/x-www-form-urlencoded')
        return resp

    def api_key_details(self, api_key, connector_id):

        resp = self.call_api('v2/apiKey/details',
                             params = {"apiKey": api_key,
                                            "connectorId": connector_id})

        return resp

    def api_key_revoke(self, connector_id):

        resp = self.call_api('v2/apiKey/revoke', method='DELETE',
                             content_type='application/x-www-form-urlencoded',
                             params = {"connectorId": connector_id})

        return resp

    def copy_token(self):
        import pyperclip
        if isinstance(self.auth, PwdAuth):
            token = self.auth._token.access_token
            pyperclip.copy(token)
            print("Copied auth token to clipboard: " + token)
        elif isinstance(self.auth, ApiKeyAuth):
            token = self.auth.api_key
            pyperclip.copy(token)
            print("Copied API Key to clipboard: " + token)
        else:
            raise Exception("Auth object not set. Can't fetch token.")





import xmltodict
import os
from ..uniform_data_rule import ExternalServiceInterface
from ..uniform_data_rule import DATA_STORAGE as BASE_DATA_STORAGE

DATA_STORAGE = os.path.join(BASE_DATA_STORAGE, 'iso')


class IsoSrvc(ExternalServiceInterface):
    """ Interface to interpret ISO XML values

    """

    @classmethod
    def get_from(cls, iso_num, field, table, **options):
        """ Get ISO information from XML file

            ISO: ISO number
            table: Table where ISO values are located
            field: Field that have ISO values
        """
        doc = cls._get_XML_from_iso(iso_num)
        return cls._to_mdmuniform_shape(doc, iso_num, field, table)

    @staticmethod
    def _get_XML_from_iso(iso_num: str):
        """ Access file or database and get corresponding XML

            iso_num: ISO number
        """
        filename = os.path.join(DATA_STORAGE, 'iso' + iso_num + '.xml')
        with open(filename) as fd:
            doc = xmltodict.parse(fd.read())
        _, doc = doc.popitem()
        return doc

    @staticmethod
    def _to_mdmuniform_shape(doc, iso_num, field, table):
        """ Convert dict of ISO into MdmUniform format

            table: Str with table name
            field: Name of field

            Return:
                List of values for specified table and field

            E.g.:

                _to_mdm_uniform_shape(doc, 'ISONUM', 'Table1', 'Field1')
                Would output:
                    {
                        'ISO_ISONUM_Field1':[]
                    }
        """
        ret = []
        items = list(list(doc.items())[1][1].items())
        for i in range(len(items)):
            table_name = items[i][0]

            if table_name == table:
                table_values = items[i][1]
                for i in table_values:
                    if field in i:
                        ret.append(i[field])
        return {f'ISO_{iso_num}': ret}


import unittest
from ..data_model import DataModel, Field
from unittest.mock import patch
from .assets import datamodel_mock
from importlib import reload


datamodelmock_carol = {'mdmName': 'datamodelmock',
                       'mdmFields': [
                           {
                               'mdmName': 'nestedone',
                               'mdmMappingDataType': 'NESTED',
                               'mdmFields': [
                                   {
                                       'mdmName': 'nestedonedeep',
                                       'mdmMappingDataType': 'LONG',
                                       'mdmFields': []
                                   }
                               ]
                           },
                           {
                               'mdmName': 'nestedtwo',
                               'mdmMappingDataType': 'DOUBLE',
                               'mdmFields': []
                           }
                       ]}


class TestDataModel(unittest.TestCase):
    def setUp(self):
        reload(datamodel_mock)
        self.Datamodelmock = datamodel_mock.Datamodelmock

    def test_get_fields(self):
        fields = self.Datamodelmock.get_fields()
        self.assertEqual(len(fields), 2)
        self.assertEqual(set(list(fields.keys())), set(['Nestedone', 'Nestedtwo']))

    def test_interpret_datamodel_defined_on_code(self):
        carol_dm = self.Datamodelmock.to_carol()
        self.assertDictEqual(carol_dm, datamodelmock_carol)

    @patch('pycarol.data_models.data_model_build.DataModelBuild.get_dm', return_value=datamodelmock_carol)
    def test_interpret_datamodel_defined_at_carol(self, mocked):
        carol_dm = DataModel.from_carol(None)
        self.assertEqual(set(carol_dm.__dict__.keys()), set(self.Datamodelmock.__dict__.keys()))

    def test_generate_validation_rules(self):
        pass


# Receives inputs[0] from an Igestion Task, returns success, logs
class TestDataModelValidation(unittest.TestCase):
    pass


from ..query import *
import pandas as pd
from ..utils.singleton import KeySingleton
from ..tasks import Tasks


class Validator(metaclass=KeySingleton):
    def __init__(self, carol):
        self.carol = carol
        self.cached_models = dict()
        self.messages = []

    def get_model_data(self, model):
        if model in self.cached_models:
            return self.cached_models[model]
        else:
            data = Query(self.carol).all(model).go().results
            df = pd.DataFrame(data)
            self.cached_models[model] = df
            return df

    def assert_non_empty(self, model=None, field=None, threshold=0.5, data=None):
        if data is None:
            assert model is not None
            print("Validating " + model + "." + field + "...")
            data = self.get_model_data(model)

        if field not in data.columns:
            self.add_warning({'code': 'MissingField', 'data-model': model, 'field': field})
            return
        non_empty = data.apply(lambda x: x.count(), axis=0)
        total = len(data)
        ratio = non_empty[field] / total
        if ratio < threshold:
            self.add_warning({'code': 'EmptyField', 'data-model': model, 'field': field,
                              'value': ratio, 'threshold': threshold})
        else:
            self.add_info({'code': 'EmptyField', 'data-model': model, 'field': field,
                              'value': ratio, 'threshold': threshold})

    def assert_min_quantity(self, model=None, min_qty=0, data=None):
        if data is None:
            assert model is not None
            print("Validating " + model + "...")
            data = self.get_model_data(model)

        total = len(data)
        if total < min_qty:
            self.add_warning({'code': 'ModelQuantity', 'data-model': model,
                              'value': total, 'threshold': min_qty})
        else:
            self.add_info({'code': 'ModelQuantity', 'data-model': model,
                           'value': total, 'threshold': min_qty})

    def assert_custom(self, code, value, threshold, not_message=None, yes_message=None):
        if value < threshold:
            self.add_warning({'code': code, 'value': value, 'threshold': threshold, 'message': not_message})
        else:
            self.add_info({'code': code, 'value': value, 'threshold': threshold, 'message': yes_message})

    def add_info(self, info):
        info['level'] = 'INFO'
        if 'code' not in info:
            info['code'] = 'Custom'
        self.messages.append(info)

    def add_warning(self, warning):
        warning['level'] = 'WARNING'
        if 'code' not in warning:
            warning['code'] = 'Custom'
        self.messages.append(warning)

    def results(self):
        return self.messages

    def post_results(self):
        tasks = Tasks(self.carol)
        for msg in self.messages:
            if msg['level'] == 'ERROR':
                tasks.error(msg['message'])
            elif msg['level'] == 'INFO':
                tasks.info(msg['message'])
            elif msg['level'] == 'DEBUG':
                tasks.debug(msg['message'])
            elif msg['level'] == 'WARNING':
                tasks.warn(msg['message'])




def matches(data, pattern):
    raise NotImplementedError
    # TODO


def contains(data, value):
    raise NotImplementedError
    # TODO


def contains_array(data, value_array):
    raise NotImplementedError
    # TODO


def starts_with(data, value):
    raise NotImplementedError
    # TODO


def ends_with(data, value):
    raise NotImplementedError
    # TODO


def equals(data, value):
    if isinstance(value, list):
        return not any(data not in value)
    return not any(data != value)


def is_empty(data, value):
    raise NotImplementedError
    # TODO


def is_email(data, value):
    raise NotImplementedError
    # TODO


def length_compare(data, value):
    raise NotImplementedError
    # TODO


def length_range(data, value):
    raise NotImplementedError
    # TODO


def is_not_number_only(data):
    return all([not d.isdigit() for d in data])


def is_invalid_brazil_state_tax_id(data):
    raise NotImplementedError
    # TODO


def is_invalid_brazil_tax_id(data):
    raise NotImplementedError
    # TODO


def value_compare(data):
    raise NotImplementedError
    # TODO


def lookup(data):
    raise NotImplementedError
    # TODO


def lookup_multiple(data):
    raise NotImplementedError
    # TODO


def lookup_found(data):
    raise NotImplementedError
    # TODO




def validate_string(data):
    return all(isinstance(d, str) for d in data)


def validate_boolean(data):
    return all(isinstance(d, bool) for d in data)


def validate_date(data):
    # TODO
    return True


def validate_long(data):
    return all(isinstance(d, int) for d in data)


def validate_double(data):
    return all(isinstance(d, float) for d in data)


def validate_binary(data):
    # TODO
    return True


def validate_object(data):
    # TODO
    return True


def validate_geopoint(data):
    # TODO
    return True




from .task import(
    Task,
    WrapperTask,
    inherit_list,
    inherit_dict
)

from .targets import (
    CDSTarget,
    PickleTarget,
    KerasTarget,
    DummyTarget,
    JsonTarget,
    FeatherTarget,
    PytorchTarget,
    PickleTarget,
    PytorchTarget,
    KerasTarget,
    LocalTarget,
    PickleLocalTarget,
    KerasLocalTarget,
    DummyTarget,
    JsonLocalTarget,
    FeatherLocalTarget,
    PytorchLocalTarget,
    PicklePyCarolTarget,
    PytorchPyCarolTarget,
    KerasPyCarolTarget,
)

from .taskviewer import (
    TaskViewer,
)

from .visualization import (
    Visualization
)

# from .dockertask import (
#     EasyDockerTask
# )

# from .kubernetestask import KubernetesJobTask


import os
import logging


class CarolHandler(logging.Handler):
    """ Handler to organize Carol logging
    This implementation send logs to the Google Stackdriver.

    Required OS Environment variables:
        * GOOGLE_APPLICATION_CREDENTIALS
        * CAROLPROCESSTYPE
        * CAROLONLINENAME | CAROLBATCHNAME
        * CAROLTENANT
        * CAROLAPPNAME
        * CAROLAPPVERSION

    Usage:
        from logger import Logger
        logger = Logger()
        logger.log.warning('It`s a warning message')
        logger.log.error('It`s an error message')
        logger.log.critical('It`s a critical message')
        ## It's possible add another handler
        # logger.log.addHandler(<handle obj>)
    """

    def __init__(self, level=logging.NOTSET):
        super().__init__(level)
        process_type = os.environ.get('CAROLPROCESSTYPE', '')
        tenant = os.environ.get('CAROLTENANT', '')
        app_name = os.environ.get('CAROLAPPNAME', '')
        app_version = os.environ.get('CAROLAPPVERSION', '')
        process_name = ''
        if process_type.lower() == 'online':
            process_name = os.environ.get('CAROLONLINENAME', '')
        elif process_type.lower() == 'batch':
            process_name = os.environ.get('CAROLBATCHNAME', '')
        self.formatter = logging.Formatter('{} > {} > {} > {} > {} > %(message)s'.format(
            tenant, app_name, app_version, process_type, process_name))

        if os.environ.get('GOOGLE_APPLICATION_CREDENTIALS'):
            from google.cloud import logging as google_logging
            client = google_logging.Client()
            self.handler = client.get_default_handler()
        else:
            raise ValueError('Tried to create a CarolHandler but no credentials for a service were found.')
        self.handler.setFormatter(self.formatter)

    def emit(self, record):
        return self.handler.emit(record)


from unittest.mock import patch, MagicMock, PropertyMock
import unittest
import luigi
from pycarol.luigi_extension import Task, inherit_list
from ..luigi import SettingsDefinition, Parameter, inherits_carol
from ...query import Query
from ...carol import Carol
from ...auth import ApiKeyAuth

""" 
# Task Execution
    When testing a Task execution, there a couple of things that could be tested:
        - Execution success
        - Requires
        - Execution

# Mocks


# Special Words when defining Test Cases:

    * Test
    Every test case must start with the 'Test' keyword as default from python.unittest.
    
    * Carol
    If your test case involves Carol related tasks, use the Carol keyword. E.g. TestCarolMyTest. This way, Carol updates
    will be able to be checked. 
    
    # TODO:
        - Check significant parameters behavior
"""


class mock_carol_query:
    def __init__(self, *mock_queries):
        """
        :param mock_queries: list of dicts
        """
        self.mock_queries = mock_queries

    def __call__(self, test_func):
        """
        MOCK_QUERY: pass a query, so any query with this exact value will be mocked

        MOCK_DM_QUERY: pass a datamodel so any query with "mdmValue" = MOCK_DM_QUERY + "Golden" will be mocked.

        QUERY_RESPONSE:

        MAX_HITS: define a maximum number of hits to a query

        RESPONSE_TYPE:
        It is expected that each query response follows a compact type, i.e. that query results should be similar to the
        results output from pyCarol.query.Query.go

        If, however, one wants to have the query on the format exactly as it comes from Carol, it is necessary to set a
        parameter 'response_type' to be equal to 'carol'.

        E.g.:
        query = { ... } # Query to be mocked
        resp = { "mdmFieldAndValues":... } # Expected result as it comes from Carol

        @mock_carol_query(dict(mock_query=query, query_response=resp, response_type='carol'))
        def test....

        :param test_func:
        :return:
        """
        # first get a reference to the original unbound method we want to mock
        original_go = Query.go
        mock_queries = self.mock_queries

        # then create a wrapper whose main purpose is to record a reference to `self`
        # when it will be passed, then delegates the actual work to the unbound method
        def side_fx(self, *a, **kw):
            side_fx.self = self
            for dic in mock_queries:
                mock_query = None
                mock_dm_query = None
                if 'mock_query' in dic:
                    mock_query = dic['mock_query']
                if 'mock_dm_query' in dic:
                    mock_dm_query = dic['mock_dm_query']
                if 'query_response' in dic:
                    query_resp = dic['query_response']
                if 'max_hits' in dic:
                    self.max_hits = dic['max_hits']
                    self.get_all = False
                response_type = 'compact'  # Default value for response_type
                if 'response_type' in dic:
                    response_type = dic['response_type']

                replace_query = False
                if mock_query is not None:
                    if self.json_query == mock_query:
                        replace_query = True
                elif mock_dm_query is not None:
                    if self.json_query.get('mustList')[0].get('mdmValue') == mock_dm_query+'Golden':
                        replace_query = True

                if replace_query:
                    class QueryResponse:
                        # get mdmGoldenFieldAndValues if not empty and if it exists
                        if response_type == 'carol':
                            results = [elem.get('mdmGoldenFieldAndValues', elem)
                                       for elem in query_resp if elem.get('mdmGoldenFieldAndValues', None)]
                        elif response_type == 'compact':
                            results = query_resp

                    return QueryResponse()
            return original_go(self, *a, **kw)

        @patch('pycarol.query.Query.go', autospec=True, side_effect=side_fx)
        def mocked_test(self, patched, *args, **kwargs):
            test_func(self, *args, **kwargs)

        return mocked_test #TODO What if mocked_test has args?


def mock_carol_app(func):
    # TODO - Mock functions related to Carol App functionalities
    pass


class TaskMock(Task):
    TARGET_DIR = "app/data/test/targets/luigi_extension/task"


class TestLuigiExtensionParameterTask(unittest.TestCase):
    # def test_get_parameters_from_carol(self):
    #     expected_params = {"abandoned":'1_11,2_11',
    #                        "branch":'ALL',
    #                        "startdate":'01/01/2008'}
    #
    #     class P(SettingsDefinition):
    #         abandoned = Parameter(carol=True, carol_name='abandonedenrollment', description='')
    #         branch = Parameter(carol=True, carol_name='subsidiary', description='')
    #         startdate = Parameter(carol=True, carol_name='initialdate', description='')
    #
    #     app_config = test_config.config[TEST_APP_TENANT]
    #     P.app = {}
    #     P.app_name = app_config['APP_NAME']
    #     P.api_token = Carol(TEST_APP_TENANT,
    #                         app_config['APP_NAME'],
    #                         auth=ApiKeyAuth(app_config['X_AUTH_KEY']),
    #                         connector_id=app_config['X_CONNECTOR_ID'])
    #     set_parameters.get_app_params(P)
    #     self.assertEqual(expected_params, P.app)

    def test_task_gets_parameters_from_all_inherited_tasks_and_set_them_to_what_is_in_ParameterTask(self):

        test_list = []

        class TaskA1(luigi.Task):
            param_a1 = Parameter()
            task_complete = False

            def run(self):
                test_list.append(('a', self.param_a1))
                self.task_complete = True

            def complete(self):
                return self.task_complete

        class TaskA2(luigi.Task):
            param_a2 = Parameter()
            task_complete = False

            def run(self):
                test_list.append(('a', self.param_a2))
                self.task_complete = True

            def complete(self):
                return self.task_complete

        @inherit_list((TaskA1, dict(param_a1=4)))
        class TaskB(TaskMock):
            param_b = Parameter()
            task_complete = False

            def run(self):
                test_list.append(('b', self.param_b))
                self.task_complete = True

            def complete(self):
                return self.task_complete

        @inherit_list(TaskB)
        class TaskC(TaskMock):
            param_c = Parameter()
            task_complete = False

            def run(self):
                test_list.append(('c', self.param_c))
                self.task_complete = True

            def complete(self):
                return self.task_complete

        @inherit_list(TaskA2)
        class TaskD(TaskMock):
            param_c = Parameter()
            task_complete = False

            def run(self):
                test_list.append(('d', self.param_c))
                self.task_complete = True

            def complete(self):
                return self.task_complete

        class AppParams(SettingsDefinition):
            TARGET_DIR = "app/data/test/targets/luigi_extension"
            param_a1 = Parameter(default=1)
            param_a2 = Parameter(default=2)
            param_b = Parameter(default=3)
            param_c = Parameter(default=5)

        AppParams.api_token = None

        @inherits_carol(AppParams)
        class Run(luigi.WrapperTask):
            task = Parameter()

            def requires(self):
                #AppParams.copy_tasks_parameters(Run)
                return [self.clone(TaskC), self.clone(TaskD)]

        task_exec = luigi.build([Run(task='test')], local_scheduler=True)
        self.assertTrue(task_exec)
        self.assertEqual(set(test_list), set([('a', 4), ('b', 3), ('c', 5), ('a', 2), ('d', 5)]))


import unittest


class TestValidateCarolValidationRules(unittest.TestCase):
    pass


class TestValidateTypes(unittest.TestCase):
    def test_validate_string(self):
        # TODO
        pass

    def test_validate_boolean(self):
        # TODO
        pass

    def test_validate_date(self):
        # TODO
        pass

    def test_validate_long(self):
        # TODO
        pass

    def test_validate_double(self):
        # TODO
        pass

    def test_validate_binary(self):
        # TODO
        pass

    def test_validate_object(self):
        # TODO
        pass

    def test_validate_geopoint(self):
        # TODO
        pass



import itertools
import os
import json
import numpy as np

from .validation_rule import ValidationRule


DATA_STORAGE = os.path.join(os.path.dirname(__file__), 'service', 'data')
RULES_STORAGE = os.path.join(os.path.dirname(__file__), 'rules')


class MdmUniformType:
    """ Data type for uniformity on MDM Fields

    """

    def __init__(self, name, description='', version='1.0.0', is_global=True, values=None, formats=None, id=None,
                 **kwargs):
        """
            name: Identifier label
            last_update: date
            description: Description of rule
            additional_info: Dict with additional information
            version: Version for the type
            is_global: Whether is a global type or not
            values: dict of values (list)
                We chose to have a dict of lists, instead of only a list
                because it gives the freedom to compose rules from different
                sources.
        """

        if id is None:
            id = self._generate_id(name)
        self.id = id
        self.name = name
        self.description = description
        self.is_global = is_global
        self.values = values if values is not None else {}
        self.formats = formats if formats is not None else {}

    @staticmethod
    def _generate_id(name):
        """ Generate a unique id for the rule
        """
        # TODO: Define a way to generate ids
        return name + 'Id'

    def post(self):
        """ Store rule

            TODO: Adapt to Carol Infrastructure
        """
        rule_dict = self.to_json()
        filename = self.id + '.json'
        filename = os.path.join(RULES_STORAGE, filename)
        with open(filename, 'w') as outfile:
            json.dump(rule_dict, outfile)
        return self

    @staticmethod
    def from_id(id):
        """ Return object with the provided id

        :param id:
        :return:
        """
        # TODO
        pass

    def get_id(self):
        return self.id

    @staticmethod
    def get_rule(id, tenant_id=None, auth=None):
        """ Get rule by id

        If tenant is None, rule is global

        :param id: Rule id
        :param tenant_id: If tenant_id is None, rule is global, otherwise get rule from tenant
        :param auth: Authentication to access tenant, if applicable
        :return: MdmUniformType rule
        """
        # TODO: Change that to a service of rules, or other thing
        # List rules storage
        try:
            filename = id+'.json'
            with open(os.path.join(RULES_STORAGE, filename), 'r') as fp:
                dic = json.load(fp)
                return MdmUniformType(**dic)
        except OSError:
            raise ValueError(f'Could not open rule with ID = {id}')

    def add_values(self, values):
        # TODO Check values format
        self.values.update(values)

    def get_values_from(self, source, field, service=None, table=None, **options):
        """ Get rule values from Carol Rules

            source: Source Service or string source name (JSON Filename)
            field: field to get values
            table: if values are organized in tables
            service: If data comes from external services
        """
        if service is None:
            filename = os.path.join(DATA_STORAGE, source + '.json')
            with open(filename) as fd:
                json_dict = json.load(fd)
                self.values.update({f'{source}_{field}': json_dict[field]})
        else:
            values = service.get_from(source, field, table, **options)
            self.values.update(values)
        return self

    def get_all_values(self):
        return list(itertools.chain.from_iterable(self.values.values()))

    def create_regex(self):
        """ Create a regex rule from Mdm

            TEMP: https://trello.com/c/FUEpRgLh/198-data-model-enum-field-type
        """
        rule_values = list(self.get_all_values())
        rule_values = [(str(r), r)[isinstance(r, str)] for r in rule_values]
        regex = '^(?!' + '$|'.join(rule_values) + '$).+'
        return regex

    def to_json(self):
        # TODO
        return self.__dict__

    def to_validation_rule(self):
        params = self.create_regex()
        return ValidationRule(ValidationRule.Operation.MATCHES, [params])

    def validate(self, data, ignore_errors = True):
        allowed_vals = self.get_all_values()

        # Ignore null data
        data = filter(lambda v: v==v, data)

        # Success Message
        msg = self.name + ' - ' + self.description
        valid = np.array([d in allowed_vals for d in data])
        success = all(valid)
        if not success:
            invalid_data = data[~valid]
            msg = 'Wrong data: ' + str(invalid_data[:min(100, len(invalid_data))])
        return success, {'success': success, 'msg': msg}


class ExternalServiceInterface:
    """ TODO: Interface to deal with external services data sources
    """

    @classmethod
    def get_from(cls, service_name, field, table=None, **options):
        pass

    @classmethod
    def update_values(cls, service_name):
        pass



import importlib.util

# Dynamically load modules

# This part was created to handle different mappings loading. Depending on the specified mapping, the correct module
# will be dynamically loaded.

loaded_modules = {}

app_mappings = dict(
)

def _load_module(module_name, modules_path):
    spec = importlib.import_module('.'.join([modules_path, module_name]))
    loaded_modules[module_name] = spec


def _load_mapping_module(module_name, modules_path):
    global loaded_modules
    global app_mappings
    if module_name not in loaded_modules:
        _load_module(module_name, modules_path)
    module = loaded_modules[module_name]
    app_mappings[module_name] = module.get_mapping_task


def get_ingestion_task(mapping_id='mdm', dm_name='', modules_path='app.datamodel.mappings'):
    global app_mappings

    if mapping_id not in app_mappings:
        _load_mapping_module(mapping_id, modules_path)

    return app_mappings[mapping_id](dm_name)



from .utils.singleton import KeySingleton
import redis


class Cache(metaclass=KeySingleton):
    def __init__(self, carol):
        self.carol = carol
        self.r = redis.StrictRedis(host='127.0.0.1', port=6379, db=0)

    def set(self, key, val):
        key = '{}-{}-{}'.format(self.carol.tenant['mdmId'], self.carol.app_name, key)
        self.r.set(key, val)

    def get(self, key):
        key = '{}-{}-{}'.format(self.carol.tenant['mdmId'], self.carol.app_name, key)
        return self.r.get(key)



class Cloner:
    def __init__(self, carol):
        self.domain = carol.domain
        self.app_name = carol.app_name
        self.port = carol.port
        self.verbose = carol.verbose
        self.connector_id = carol.connector_id
        self.auth = carol.auth.cloner()

    def build(self):
        from .carol import Carol
        return Carol(self.domain, self.app_name, self.auth.build(), connector_id=None, port=self.port, verbose=self.verbose)



from distutils.core import setup
from setuptools import setup, find_packages

import codecs
import os
import re
import sys

here = os.path.abspath(os.path.dirname(__file__))

try:
    with open('requirements.txt', 'r') as req_file:
        requirements = req_file.read()
    install_requires = requirements.splitlines()
except FileNotFoundError as e:
    install_requires = [
        'bokeh',
        'pykube',
        'boto3==1.9.16',
        'botocore==1.12.16',
        'click==7.0',
        'dask[complete]==0.19.3',
        'toolz',
        'fastparquet',
        'flask==1.0.2',
        'google-api-core',
        'google-api-python-client',
        'google-auth-httplib2',
        'google-auth',
        'google-cloud-core',
        'google-cloud-logging',
        'google-cloud-storage',
        'googleapis-common-protos',
        'joblib==0.11',
        'luigi',
        'numpy==1.16.3',
        'pandas==0.23.4',
        'pyarrow==0.11.1',
        'redis==2.10.6',
        'requests',
        's3fs==0.1.6',
        's3transfer==0.1.13',
        'tqdm==4.28.1',
        'urllib3',
        'websocket-client==0.53.0',
        'missingno==0.4.1',
        'deprecated'
    ]

def read(*parts):
    # intentionally *not* adding an encoding option to open, See:
    #   https://github.com/pypa/virtualenv/issues/201#issuecomment-3145690
    with codecs.open(os.path.join(here, *parts), 'r') as fp:
        return fp.read()

def find_version(*file_paths):
    version_file = read(*file_paths)
    version_match = re.search(
        r"^__version__ = ['\"]([^'\"]*)['\"]",
        version_file,
        re.M,
    )
    if version_match:
        return version_match.group(1)

    raise RuntimeError("Unable to find version string.")

setup(
    name='pycarol',
    packages=find_packages(exclude=['docs', 'doc']),
    version=find_version("pycarol", "__init__.py"),
    license='TOTVS',
    description='Carol Python API and Tools',
    author='TotvsLabs',
    maintainer='TOTVS Labs',
    author_email='ops@totvslabs.com',
    url='https://www.carol.ai/',
    keywords=['Totvs', 'Carol.ai', 'AI'],
    install_requires=install_requires,
    classifiers=[
        # Chose either "3 - Alpha", "4 - Beta" or "5 - Production/Stable" as the current state of your package
        'Development Status :: 5 - Production/Stable',
        # Define that your audience are developers
        'Intended Audience :: Developers',
        'Topic :: Software Development :: Build Tools',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3.6',
        "Operating System :: OS Independent",
    ],
)



import numpy as np
import pandas as pd

class TaskViewer(object):
    nodes = []
    selected_nodes = []

    def __init__(self, task, parallel_update=True):
        self.dag = dict()
        # TaskViewer.nodes = []
        self.node_level = dict()
        self.task = task
        self.tree = self.get_tree(task)
        self.max_level = max({self.node_level[k] for k in self.node_level.keys()})
        self.root_nodes = [i for i, v in self.node_level.items() if v == self.max_level]
        self.dag_node_level = [0 for i in self.node_level]
        for i in self.range_nodes():
            self.set_dag_node_level(i, 0)
        self.rev_dag = self.get_reverse_dag()
        self.parallel_update = parallel_update

    def range_nodes(self):
        return range(len(TaskViewer.nodes))

    def get_tree(self, t, level=0):
        if t not in TaskViewer.nodes:
            TaskViewer.nodes.append(t)
        t_i = TaskViewer.nodes.index(t)
        if t_i not in self.node_level:
            self.node_level[t_i] = level
        sons = t.requires()
        for son in sons:
            if son not in TaskViewer.nodes:
                TaskViewer.nodes.append(son)
            son_i = TaskViewer.nodes.index(son)
            if son_i not in self.dag:
                self.dag[son_i] = {t_i}
            else:  # son already in dag
                self.dag[son_i].add(t_i)
            self.get_tree(son, level + 1)


    def set_dag_node_level(self, node, level=0):
        self.dag_node_level[node] = max(level, self.dag_node_level[node])
        if node not in self.dag:
            return
        for n_i in self.dag[node]:
            self.set_dag_node_level(n_i, level + 1)

    def get_reverse_dag(self):
        rd = {k: [] for k in self.range_nodes()}
        for source_node, v in self.dag.items():
            for target_node in v:
                rd[target_node].append(source_node)
        return rd

    def get_layout(self):
        x = []
        y = []
        temp_level = dict()
        node_level = self.dag_node_level
        for i in self.range_nodes():
            _x = node_level[i]
            x.append(_x)
            if node_level[i] not in temp_level:
                temp_level[node_level[i]] = 0
            else:
                temp_level[node_level[i]] += 1
            _y = temp_level[node_level[i]] + ((-_x) % 2) / 2. + np.random.random() * 0.5
            y.append(_y)
        return x, y

    def get_edges(self, x, y):
        edges_x = []
        edges_y = []
        rev_dag = self.get_reverse_dag()

        for source_node in self.range_nodes():
            l = [target_node for target_node in rev_dag[source_node]]
            result = [source_node]
            for e in l:
                result.append(e)
                result.append(source_node)
            edges_x.append([x[r] for r in result])
            edges_y.append([y[r] for r in result])
        return edges_x, edges_y

    def get_source_dict(self):
        x, y = self.get_layout()
        edges_x, edges_y = self.get_edges(x, y)

        task_names = []
        task_id = []
        for i in self.range_nodes():
            task_names.append(TaskViewer.nodes[i].task_family)
            task_id.append(TaskViewer.nodes[i].task_id)
        family = [t.split('.')[-2] if '.' in t else "empty_namespace" for t in task_names]
        task_names = [t.split('.')[-1] for t in task_names]
        self.family = family
        self.update_complete()
        source_dict = dict(
            x=x,
            y=y,
            task_names=task_names,
            family=family,
            task_id=task_id,
            edges_x=edges_x,
            edges_y=edges_y,
            complete=self.complete,
            tasklog=self.tasklog,
        )
        return source_dict

    def update_complete(self):
        import concurrent
        if self.parallel_update:
            with concurrent.futures.ThreadPoolExecutor(max_workers=5) as executor:
                complete = executor.map(lambda x: x.complete(), TaskViewer.nodes)
                tasklog = executor.map(lambda x: x.loadlog(), TaskViewer.nodes)
            self.complete = list(complete)
            self.tasklog = list(tasklog)

        else:
            self.complete = [t.complete() for t in TaskViewer.nodes ]
            self.tasklog = [t.loadlog()  for t in TaskViewer.nodes ]
        self.tasklog = [log if isinstance(log,str) else "Log type is wrong." for log in self.tasklog]
        return self.complete

    def get_colormapper(self):
        family = self.family
        factors = list(set(family))
        nb_factors = np.clip(len(factors),3,10)
        from bokeh.palettes import Category10
        from bokeh.models.mappers import CategoricalColorMapper
        return CategoricalColorMapper(factors=factors, palette=Category10[nb_factors])

    def plot_task(self, doc,web_app=False):
        # plot dashboard
        print('start plot_task')
        doc.clear()
        from bokeh.models import Circle, Text, Button
        from bokeh.models import ColumnDataSource, CDSView, BooleanFilter, HoverTool
        from bokeh.plotting import figure
        from bokeh.layouts import column, row, layout
        from bokeh.transform import transform

        source_dict = self.get_source_dict()
        cm = self.get_colormapper()

        TOOLTIPS = [
            ('index', "$index"),
            ('task_id', '@task_id'),
        ]
        hover = HoverTool(names=['alltasks'])
        plot = figure(
            title="Pipeline Debugger",
            x_range=(-1, max(source_dict['x']) + 1),
            y_range=(-1, max(source_dict['y']) + 1),
            tools=[hover, 'wheel_zoom', 'pan', 'tap', 'box_select', 'reset'],
            toolbar_location='right',
            tooltips=TOOLTIPS,
            plot_width=900,
            plot_height=500,
        )

        s1 = ColumnDataSource(data=source_dict)
        s1_edges = ColumnDataSource(data={k:v for k,v in source_dict.items() if 'edges' in k})
        complete = self.update_complete()
#         print(complete)
        s2 = ColumnDataSource(data={k: [vi for i, vi in enumerate(v) if not complete[i]] for k, v in
                                    source_dict.items()})  # select only the rows of source_dict where complete==False
        done_tasks = CDSView(source=s1, filters=[BooleanFilter(complete)])
        notdone_tasks = CDSView(source=s1, filters=[BooleanFilter([not c for c in complete])])

        ml = plot.multi_line(xs='edges_x', ys='edges_y', source=s1_edges, color='gray',line_alpha=0.3)
        ml.nonselection_glyph = None
        plot.circle('x', 'y', source=s1, size=20, name='alltasks', color=transform('family', cm), legend='family')
        plot.circle('x', 'y', source=s2, size=10, color='white')
        plot.text(x='x', y='y', text='task_names', source=s1, text_font_size='8pt')

        from bokeh.events import ButtonClick
        from bokeh.models import Button
        from bokeh.layouts import widgetbox
        from bokeh.models.widgets import PreText
        pre = PreText(text="")

        def select_callback(attr, old, new):
            TaskViewer.selected_nodes = [TaskViewer.nodes[i] for i in new]
            pre.text = self.tasklog[new[0]]
            print("selected {}".format(new))

        s1.selected.on_change('indices', select_callback)


        remove_button = Button(label='Remove Selected')

        def remove_callback(event):
            target_list = s1.selected.indices
            for t in target_list:
                print('removing', t)
                TaskViewer.nodes[t].remove()

        remove_button.on_event(ButtonClick, remove_callback)


        removeupstream_button = Button(label='Remove Upstream')

        def removeupstream_callback(event):
            def _removeupstream(t, dag):
                TaskViewer.nodes[t].remove()
                for ti in dag[t]:
                    _removeupstream(ti, dag)

            dag = self.dag
            dag.update({0: {}})
            target_list = s1.selected.indices
            for t in target_list:
                _removeupstream(t, dag)
            return

        removeupstream_button.on_event(ButtonClick, removeupstream_callback)


        update_button = Button(label='Update')

        def update_callback(event):
            complete = self.update_complete()
            s2.data = {k: [vi for i, vi in enumerate(v) if not complete[i]] for k, v in source_dict.items()}
            return

        update_button.on_event(ButtonClick, update_callback)

        if web_app:
            l = column(
                [
                    plot,
                    row(
                        remove_button,
                        removeupstream_button,
                        update_button,
                    ),
                    widgetbox(pre),
                ],
                sizing_mode='scale_width'
            )
        else:
            l = column(
                children=[
                    plot,
                    row(
                        remove_button,
                        removeupstream_button,
                        update_button,
                    )
                ],
            )

        doc.add_root(l)

    def show(self, url="http://127.0.0.1", jupyter_notebook_port=8888, bokeh_port=5006):
        from bokeh.io.state import curstate
        from bokeh.io.notebook import show_app
        from bokeh.io import output_notebook

        output_notebook()

        notebook_url = ":".join([url, str(jupyter_notebook_port)])
        return show_app(self.plot_task, curstate(), notebook_url=notebook_url, port=bokeh_port)




""" Interpret validation log to generate an HTML report

Input Format:
    {
        datamodel:STR,
        domain:STR,
        number_of_entries:INT,
        [validation_rule_name]:{
            success: BOOL,
            msg: STR
        },
        log:{
            fields:{
                [field_name]:{
                    found: BOOL,
                    [validation_rule_name]:{
                        success: BOOL
                        msg: STR
                    }
                    type:{
                        success: BOOL
                    }
                    *nested:{}
                }
            }
        }
    }
"""
import luigi
from ...luigi_extension.test import task_execution_debug
import missingno as msno
import matplotlib.pyplot as plt

from IPython import get_ipython
from IPython.display import display, Markdown, Javascript
import logging


# TODO: Clickable buttons that execute the function
SUCCESS_COLOR = "green"
WARNING_COLOR = "orange"
ERROR_COLOR = "red"
NULL_COLOR = "gray"


def insert_cell(text=''):
    get_ipython().set_next_input(text)


def get_tasks_with(execution, substring):
    val_tasks = {}
    for task in execution.task_history:
        if task[1] == 'DONE':
            task_name = task[0].__class__.__name__
            if substring in task_name:
                val_tasks.update({task_name: task[0]})
    return val_tasks


def add_tag(msg, color=None, tag=None, tag_options=None, style=None):
    if color is not None:
        msg = f'<font color="{color}">{msg}</font>'
    if tag is not None:
        if tag_options is None:
            tag_options = {}
        if color is not None:
            tag_options.update({'color': color})
        if style is not None:
            style = ";".join([f'{k}:{v}' for k, v in style.items()])
            style = f'style="{style};"'
        tag_options = " ".join([f'{k}="{v}"' for k, v in tag_options.items()])
        msg = f'<{tag} {tag_options} {style}>{msg}</{tag}>'
    return msg


def print_m(msg, color=None, tag=None, tag_options=''):
    msg = add_tag(msg, color, tag, tag_options)
    display(Markdown(msg))


def print_null_data(df, columns=None):
    if columns is not None:
        df = df[columns]
    # TODO: Break columns (max = 20)
    msno.matrix(df=df)
    display(plt.show())


def print_dm_validation(logs, data):
    # TODO: Organize by fields
    log = logs['log']

    def get_validation_text(log, data, field_name, color=SUCCESS_COLOR):
        text = '<ul>'
        if 'found' in log:
            val = log.pop('found')
            if not val:
                if log['mandatory']:
                    color = ERROR_COLOR
                    text += add_tag('Mandatory field not found.', color=ERROR_COLOR, tag='li')
                else:
                    color = WARNING_COLOR
                    text += add_tag('Field not found.', color=WARNING_COLOR, tag='li')
        if 'type' in log:
            val = log.pop('type')
            if not val['success']:
                color = ERROR_COLOR
                text += add_tag('Wrong type.', color=ERROR_COLOR, tag='li')

        if 'validation' in log:
            val = log.pop('validation')[0]
            if not val['success']:
                color = ERROR_COLOR
                text += add_tag('Error with validation rule.', color=ERROR_COLOR, tag='li')
                text += add_tag('<ul>')
                text += add_tag(val['msg'], color=ERROR_COLOR, tag='li')
                text += add_tag('</ul>')

        if 'uniformity' in log:
            val = log.pop('uniformity')
            if not val['success']:
                color = ERROR_COLOR
                text += add_tag('Error with uniformity of data.', color=ERROR_COLOR, tag='li')
                text += add_tag('<ul>')
                text += add_tag(val['msg'], color=ERROR_COLOR, tag='li')
                text += add_tag('</ul>')

        if 'custom_validation' in log:
            val = log.pop('custom_validation')
            if not val['success']:
                color = ERROR_COLOR
                text += add_tag('Error with custom validation rule.', color=ERROR_COLOR, tag='li')
                text += add_tag('<ul>')
                text += add_tag(val['msg'], color=ERROR_COLOR, tag='li')
                text += add_tag('</ul>')

        text += add_tag('</ul>')
        text = '' if text == '<ul></ul>' else text

        return text, color

    def print_field(log, data, calling_field_name=None, color=NULL_COLOR):
        text = '<ul>'
        for field_name, field in log['fields'].items():
            color = SUCCESS_COLOR
            field_name = field_name.lower()  # Get Carol field name
            aux_field_name = calling_field_name + '_' + field_name if calling_field_name is not None else field_name
            if 'nested' in field:
                validation_text, color = print_field(field['nested'], data, aux_field_name, color)
            else:
                validation_text, color = get_validation_text(field.copy(), data=data, field_name=aux_field_name,
                                                             color=color)

            text += add_tag(add_tag(field_name, style={'background-color': color}, tag='button'), tag='li')
            text += validation_text

        text += '</ul>'

        return text, color

    text, _ = print_field(log, data)
    print_m(text)


def display_dm_report(val, dm_name, ingestion=True):
    """
        val = Luigi Extension Validation Task
        dm_name = str DataModel name
    """
    logs = val.load()
    # TODO: List of data models
    print_m(f'## {logs["datamodel"].capitalize()} '
            f'Data Model <a id="{dm_name}"></a>')
    df = val.requires()[0].load()
    print_m('### Validation:')
    print_dm_validation(val.load(), data=df[0:10])
    if ingestion:
        print_m(f'### Ingestion')
        print_m(f'#### Shape: {df.shape}')
        display(df.head())
        print_m('### Null data:')
        print_null_data(df)
    del df


def display_report(execution):
    print_m(f"# Carol App Validation - {execution.task.domain}")
    print_m('### Navigation')
    dm_val_tasks = get_tasks_with(execution, 'DataModelValidation')
    text = ''
    for dm in dm_val_tasks.keys():
        dm_name = dm.replace("DataModelValidation", "")
        text += f'* [{dm_name}](#{dm})\n'
    print_m(text)
    for val_task_name, val in dm_val_tasks.items():
        val = dm_val_tasks[val_task_name]
        display_dm_report(val, val_task_name)
    return dm_val_tasks


def generate_report(pipeline_task, domain=None, dm='all', mute_luigi=True):
    print('Running luigi validation pipeline...')
    if mute_luigi:
        luigi.interface.setup_interface_logging.has_run = True
    params = dict(operation='dm_validation', task=dm)
    if domain is not None:
        params.update({'domain': domain})
    execution = task_execution_debug(pipeline_task, params)
    if not execution.success:
        print('Error with pipeline.')
    # TODO: Check if any task was not suceeded

    return display_report(execution)



"""

"""
import inspect
import sys
import numpy as np
from .validation_rule.validation_rule import ValidationRule
from .validation_rule.uniform_data_rule import MdmUniformType
from .validation_rule.validation_funcs import type_rules as type_rules_funcs

# TODO: Relation: if this field is related to another data model
# TODO: Type: Default value + Type custom validations


class Field:
    """ Manages Carol Field

        * validation_rule
        * uniform_rule
        * custom_validation (data)

    """
    _carol_name = None  # Field name at Carol. If None, will return Class's __name__

    _mandatory = False  # If field is mandatory or not
    _min_fill_perc = 0  # Minimum number of non null values

    def set_foreign_key(self):
        """ Set this field as a foreign key. Must return the field related to this key.
        :return:
        """
        return None

    class TYPE:
        STRING = "STRING"
        NESTED = "NESTED"
        BOOLEAN = "BOOLEAN"
        DATE = "DATE"
        LONG = "LONG"
        DOUBLE = "DOUBLE"
        BINARY = "BASE64"
        OBJECT = "OBJECT"
        GEOPOINT = "GEOPOINT"

    @classmethod
    def get_name(cls):
        """ Class name should be equal to field name. If not the case, overload this method to return field name.

            E.g. Field with name 'fieldname' should have a class named like:
                class FieldName(DataModel):
                    ...
        """
        if cls._carol_name is None:
            return cls.__name__.lower()
        else:
            return cls._carol_name

    @classmethod
    def get_fields(cls):
        """ Get Nested Fields from this class

        :return:
        """
        fields = {}
        for att_name, att_val in cls.__dict__.items():
            if inspect.isclass(att_val):
                if issubclass(att_val, Field):
                    fields.update({att_name: att_val})
        return fields

    @classmethod
    def get_validation_rules(cls):
        validations = {}
        for att_name in dir(cls):
            if att_name in ['validation_rule', 'uniform_rule', 'custom_validation']:
                validations.update({att_name: getattr(cls, att_name)})
        return validations

    @classmethod
    def add_field(cls, field):
        """ Add Field to DataModel

        :param field: Field
        :return:
        """
        setattr(cls, field.__name__, field)

    @classmethod
    def from_carol(cls, field_info, validation_info=None):
        """ Interpret data from Carol json info of a field

        :param field_info: dict with information of a field following Carol MDM Model
        :return: new Field class
        """
        # Transform all fields into classes, and all nested fields into subclasses
        field_name = field_info['mdmName'].capitalize()
        field_type = field_info['mdmMappingDataType']
        field_class = type(field_name, (Field,), {'TYPE': field_type})

        if field_type == Field.TYPE.NESTED:
            for nested_field in field_info['mdmFields']:
                nested_field_name = nested_field['mdmName'].capitalize()
                nested_field_class = Field.from_carol(nested_field)
                setattr(field_class, nested_field_name, nested_field_class)

        if validation_info is not None:
            # TODO: Add field
            pass

        return field_class

    @classmethod
    def to_carol(cls, add_metadata=False, force_create=False):
        """ Convert from Field class format to Carol dict

        :return: Carol dict
        """
        field_carol = {'mdmName': cls.get_name(),
                       'mdmMappingDataType': cls.TYPE,
                       'mdmFields': []}

        for att_name, att_val in cls.__dict__.items():
            if inspect.isclass(att_val):
                if issubclass(att_val, Field):
                    field_carol['mdmFields'].append(att_val.to_carol(add_metadata=add_metadata,
                                                                     force_create=force_create))
        return field_carol

    @classmethod
    def validate(cls, data, ignore_errors: bool=False, _nested=None, validate_nested=True):
        """ Validate if data follows validation definition

        :param data: Pandas data frame with data
        :param ignore_errors: ignore errors and continue (True) stop execution when validation error found (False)
        :param validate_nested: if True, will continue to get validation from nested fields. If False,
            logs['nested'] will return dict of fields with the format: {  field_name : (field_class, _nested) }
        :return: success (Boolean), logs (dict)
        """
        success = True
        logs = {}
        logs.update({'success': False})
        logs.update({'mandatory': cls._mandatory})
        if _nested is None:
            _nested = []
        _nested = _nested.copy()  # Python crazy way of dealing with lists
        _nested.append(cls.get_name())

        if cls.TYPE != Field.TYPE.NESTED:
            field_key = '_'.join(_nested)
            # TODO: What if fields weren't supposed to come necessarily?
            if field_key not in data.columns:
                if not ignore_errors:
                    raise ValueError(f'Could not find field {cls.get_name()} in data.')
                logs.update({'found': False})
                if cls._mandatory:
                    success = False

            else:
                logs.update({'found': True})
                data = data[field_key].dropna()
                if len(data) == 0:
                    logs.update({'non_null': False})
                    return False, logs

                # Minimum Null Validation
                if cls._min_fill_perc > 0:
                    # TODO
                    pass

                # Type Validation
                type_fun = getattr(type_rules_funcs, 'validate_' + cls.TYPE.lower())  # Get type validation function
                type_success = type_fun(data)
                logs.update({'type': {'success': type_success}})
                if not type_success:
                    if not ignore_errors:
                        raise ValueError(f'Error when validating type of field: {cls.get_name()}')
                    return type_success, logs
                success = success and type_success

                # Carol Validation Rule
                if hasattr(cls, 'validation_rule'):
                    validation_rule = cls.validation_rule()
                    validation_logs = []
                    if not isinstance(validation_rule, list):
                        validation_rule = [validation_rule]
                    for val in validation_rule:
                        if not isinstance(val, ValidationRule):
                            raise TypeError('validation_rule must return a ValidationRule object.')
                        try:
                            validation_success, val_logs = val.validate(data)
                        except TypeError as e:
                            validation_success = False
                            val_logs = str(e)
                        if not ignore_errors and not validation_success:
                            raise ValueError(f'Error when validating field [{cls.get_name()}]')
                        success = success and validation_success
                        validation_logs.append(val_logs)
                    logs.update({'validation': validation_logs})

                # Uniform_Rule Validation
                if hasattr(cls, 'uniform_rule'):
                    uniform_rule = cls.uniform_rule()
                    if not isinstance(uniform_rule, MdmUniformType):
                        raise TypeError('uniform_rule must return an MdmUniformType object.')
                    try:
                        val_success, val_logs = uniform_rule.validate(data)
                    except TypeError as e:
                        val_success = False
                        val_logs = str(e)
                    if not ignore_errors and not val_success:
                        raise ValueError(f'Error when validating uniformity of field [{cls.get_name()}]')
                    success = success and val_success
                    logs.update({'uniformity': val_logs})

                # Custom Validation
                if hasattr(cls, 'custom_validation'):
                    # TODO Add msg from function's docstring
                    custom_val_msg = 'Custom Validation - Check function code for more details.'
                    try:
                        custom_val_success = cls.custom_validation(data)
                    except TypeError as e:
                        custom_val_success = False
                        custom_val_msg = str(e)
                    success = success and custom_val_success
                    logs.update({'custom_validation': {'success': custom_val_success,
                                                       'msg': custom_val_msg}})


        else:
            # Validate nested
            logs['nested'] = {}
            logs['nested']['fields'] = {}
            for field_name, field in cls.get_fields().items():
                if validate_nested:
                    f_success, f_logs = field.validate(data, ignore_errors=ignore_errors, _nested=_nested)
                    success = success and f_success
                    logs['nested']['fields'].update({field_name: f_logs})
                else:
                    logs['nested']['fields'].update({field_name: (field, _nested)})

        logs.update({'success': success})
        return success, logs

    @classmethod
    def get_nested_validation_rules(cls):
        """ Used to transform field validation rules to data model validation rule

        :return: List with all validation rules in a dictionary format {cls: validation_rule}
        """
        validation_rule = cls.validation_rule()

        # Get nested validation rules
        rules = [{cls: validation_rule}] if validation_rule is not None else []
        for field in cls.get_fields().values():
            rules += field.get_nested_validation_rules()
        return rules

    @classmethod
    def get_dtype(cls):
        return {Field.TYPE.BOOLEAN: '?',  # boolean
                Field.TYPE.STRING: 'O',  # (Python) objects
                Field.TYPE.NESTED: 'O',  # (Python) objects
                Field.TYPE.DATE: "M",
                Field.TYPE.LONG: "i",
                Field.TYPE.DOUBLE: "f",
                Field.TYPE.BINARY: "O",
                Field.TYPE.OBJECT: "O",
                Field.TYPE.GEOPOINT: "O"
            }[cls.TYPE]

    @classmethod
    def get_parent(cls):
        """ Get class that is the Nested Field's parent

        :return: Parent's class
        """
        parents_name = cls.__qualname__.split('.')

        def rec_get_parent(i):
            if i == 0:
                return getattr(sys.modules[cls.__module__], parents_name[0])  # Nested are attr of main class
            return getattr(rec_get_parent(i-1), parents_name[i])
        return rec_get_parent(len(parents_name)-2)



from ...data_models.data_model_build import DataModelBuild
from .validation_funcs import carol_validation_rules


class ValidationRule:
    """ Class to manage validation rules

    """

    class OPERATION:
        CUSTOM = 'CUSTOM_VALIDATION_FUNCTION',
        MATCHES = 'MATCHES'
        CONTAINS = 'CONTAINS'
        CONTAINS_ARRAY = 'CONTAINS_ARRAY'
        STARTS_WITH = 'STARTS_WITH'
        ENDS_WITH = 'ENDS_WITH'
        EQUALS = 'EQUALS'
        IS_EMPTY = 'IS_EMPTY'
        IS_EMAIL = 'IS_EMAIL'
        LENGTH_COMPARE = 'LENGTH_COMPARE'
        LENGTH_RANGE = 'LENGTH_RANGE'
        IS_NOT_NUMBER_ONLY = 'IS_NOT_NUMBER_ONLY'
        IS_INVALID_BRAZIL_STATE_TAX_ID = 'IS_INVALID_BRAZIL_STATE_TAX_ID'
        IS_INVALID_BRAZIL_TAX_ID = 'IS_INVALID_BRAZIL_TAX_ID'
        VALUE_COMPARE = 'VALUE_COMPARE'
        LOOKUP = 'LOOKUP'
        LOOKUP_MULTIPLE = 'LOOKUP_MULTIPLE'
        LOOKUP_FOUND = 'LOOKUP_FOUND'

    def __init__(self, operation:str =None, params:list=None, func=None):
        """

        :param operation:
        :param params:
        :param func: (Optional) Custom function
        """
        if func is None:
            if operation == ValidationRule.OPERATION.CUSTOM:
                raise AttributeError('If operation is custom, validation func must be provided.')
            func = getattr(carol_validation_rules, operation.lower())
        else:
            operation = ValidationRule.OPERATION.CUSTOM

        self.operation = operation
        self.params = params
        self.func = func

    @staticmethod
    def get_carol_operation_id(operation):
        # TODO
        pass

    def to_carol(self, dm_name, field_name, carol, add_to_carol=False):
        """ Get validation rule in Carol MDM's format

        :param carol: Carol object if want to add rule to Carol if not there
        :param add_to_carol: If rule does not exist on Carol, add it or not
        :return: Dict of the Validation Rule on Carol MDM Format
        """
        dmb = DataModelBuild(carol)
        body = {}

        if add_to_carol:
            rule_id, _ = dmb.create_rule(dm_name, field_name, self.operation, self.params)
            body.update({'mdmId': rule_id})

        template_id = dmb.get_dm(dm_name)['mdmId']
        body.update(dmb.generate_rule_body(template_id, dm_name, field_name, self.operation, self.params))
        return body

    @classmethod
    def from_carol(cls, val_rule):
        # TODO
        pass

    def validate(self, data):
        """

        :param data:
        :return: Success (boolean), logs (dict)
        """
        msg = self.operation

        # Remove non null data
        # TODO

        if self.params is not None:
            msg += ' - Params: ' + str(self.params)
            success = self.func(data, *self.params)
        else:
            success = self.func(data)

        return success, {'success': success, 'msg': msg}


import pandas as pd
from dask import dataframe as dd
from .. import __BUCKET_NAME__
import io
from joblib import Parallel, delayed
from tqdm import tqdm

__STAGING_FIELDS = ['mdmCounterForEntity', 'mdmId']
__DM_FIELDS = ['mdmCounterForEntity', 'mdmId']


def _import_dask(storage, merge_records=False,
                 dm_name=None, golden=False, return_dask_graph=False,
                 connector_id=None, staging_name=None, columns=None, max_hits=None, mapping_columns=None):
    if columns:
        columns = list(set(columns))
        columns += __STAGING_FIELDS
        columns = list(set(columns))

    if golden:
        url = [storage.build_url_parquet_golden(dm_name=dm_name)]
    else:
        url = []
        url1 = storage.build_url_parquet_staging(staging_name=staging_name, connector_id=connector_id)
        if url1 is not None:
            url.append(url1)

        url2 = storage.build_url_parquet_staging_master(staging_name=staging_name, connector_id=connector_id)
        if url2 is not None:
            url.append(url2)

        url3 = storage.build_url_parquet_staging_rejected(staging_name=staging_name, connector_id=connector_id)
        if url3 is not None:
            url.append(url3)

    d = dd.read_parquet(url, storage_options=storage.get_dask_options(), columns=columns)

    d= d.rename(columns=mapping_columns)
    if return_dask_graph:
        return d
    else:
        return d.compute()


def _import_pandas(storage, dm_name=None, connector_id=None, columns=None, mapping_columns=None,
                   staging_name=None, golden=False, max_hits=None, callback=None):
    if columns:
        columns = list(set(columns))
        columns += __DM_FIELDS
        columns = list(set(columns))

    if golden:
        file_paths = storage.get_golden_file_paths(dm_name=dm_name)
    else:
        file_paths = storage.get_staging_file_paths(staging_name=staging_name, connector_id=connector_id)

    df_list = []
    count = 0
    for i, file in enumerate(tqdm(file_paths)):
        buffer = storage.load(file['name'], format='raw', cache=False, storage_space=file['storage_space'])
        result = pd.read_parquet(buffer, columns=columns)

        if mapping_columns is not None:
            result.rename(columns=mapping_columns, inplace=True) #fix columns names (we replace `-` for `_` due to parquet limitations.
        if callback:
            assert callable(callback), \
                f'"{callback}" is a {type(callback)} and is not callable. This variable must be a function/class.'
            result = callback(result)

        df_list.append(result)
        if max_hits is not None:
            count_old = count
            count += len(df_list[i])
            if count >= max_hits:
                df_list[i] = df_list[i].iloc[:max_hits - count_old]
                break
    if not df_list:
        return None
    return pd.concat(df_list, ignore_index=True)
